#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri May  8 11:56:29 2020

@author: jojolinj
"""
from ase import Atoms
from ase.io import read
import numpy as np
from dscribe.descriptors import MBTR as dMBTR, SOAP
from pathlib import Path
from ase.data import atomic_numbers, chemical_symbols
#from ase.symbols import Symbols
from typing import List
import sys
from time import time
from diskOp import getFilePaths
#import re


def setup_Dscribe_MBTR(atomic_symbols, ks: int or List[int], acc_N: int,
                       flatten=True):
    """
    """
    k1=False; k2=False; k3=False
    if type(ks)==list:
        k1 = k1 or (1 in ks)
        k2 = k2 or (2 in ks)
        k3 = k3 or (3 in ks)
    else:
        k1 = k1 or (1==ks)
        k2 = k2 or (2==ks)
        k3 = k3 or (3==ks)

    K1=None; K2=None; K3=None
    
    try:
        if k1:
            K1 = {
            "geometry": {"function": "atomic_number"},
            "grid": {"min": 0, "max": 8, "n": acc_N, "sigma": 0.1}}
        if k2:
            K2 = {
            "geometry": {"function": "inverse_distance"},
            "grid": {"min": 0, "max": 1/0.7, "n": acc_N, "sigma": 0.1},
            "weighting": {"function": "exponential", "scale": 0.5, "cutoff": 1e-3}}
        if k3:
            K3 = {
            "geometry": {"function": "cosine"},
            "grid": {"min": -1, "max": 1, "n": acc_N, "sigma": 0.1},
            "weighting": {"function": "exponential", "scale": 0.5, "cutoff": 1e-3}}
    
        return dMBTR(species=atomic_symbols,
                     k1=K1,
                     k2=K2,
                     k3=K3,
                     periodic=False,
                     flatten=flatten,
                     normalization="l2_each")
    except ValueError:
        if k1:
            K1 = {
            "geometry": {"function": "atomic_number"},
            "grid": {"min": 0, "max": 8, "n": acc_N, "sigma": 0.1}}
        if k2:
            K2 = {
            "geometry": {"function": "inverse_distance"},
            "grid": {"min": 0, "max": 1/0.7, "n": acc_N, "sigma": 0.1},
            "weighting": {"function": "exp", "scale": 0.5, "threshold": 1e-3}}
        if k3:
            K3 = {
            "geometry": {"function": "cosine"},
            "grid": {"min": -1, "max": 1, "n": acc_N, "sigma": 0.1},
            "weighting": {"function": "exp", "scale": 0.5, "threshold": 1e-3}}
    
        return dMBTR(species=atomic_symbols,
                     k1=K1,
                     k2=K2,
                     k3=K3,
                     periodic=False,
                     flatten=flatten,
                     normalization="l2_each")
    

def setup_Dscribe_SOAP(species, rcut, n_max=7, l_max=4, sigma=0.25):
    """
    """
    return SOAP(
        species=species,
        periodic=False,
        rcut=rcut,
        nmax=n_max,
        lmax=l_max,
        sigma=sigma,
        )


def calculate_Dscribe_mbtrs(trajs: List[Atoms], mbtr: dMBTR,
                            N_as: int, acc_N: int,
                            ks: int or List[int], flatten=True) -> dict:
    """
    """

    if flatten:
        mbtrs = mbtr.create(trajs[:],n_jobs=8)
        return mbtrs

    if not flatten:
        mbtrs = mbtr.create(trajs[:],n_jobs=8)
        return mbtrs

    def nts(num: int):
        return 'k{0:d}'.format(num)


    N_clu = len(trajs)

    mbtrs = {}
    assert type(ks) == int or type(ks) == List


    # Initialization
    if type(ks) == int:
        if ks==1:
            mbtrs[nts(ks)] = np.zeros((N_clu, N_as, acc_N),
                 dtype=np.float)
        elif ks==2:
            mbtrs[nts(ks)] = np.zeros((N_clu, N_as, N_as, acc_N),
                 dtype=np.float)
        elif ks==3:
            mbtrs[nts(ks)] = np.zeros((N_clu, N_as, N_as, N_as, acc_N),
                 dtype=np.float)
        else:
            raise NotImplementedError
        ksl = [nts(ks)]

    else:
        ksl = []
        if 1 in ks:
            mbtrs['k1'] = np.zeros((N_clu, N_as, acc_N), dtype=np.float)
            ksl.append('k1')

        if 2 in ks:
            mbtrs['k2'] = np.zeros((N_clu, N_as, N_as, acc_N), dtype=np.float)
            ksl.append('k2')

        if 3 in ks:
            mbtrs['k3'] = np.zeros((N_clu, N_as, N_as, N_as, acc_N),
                 dtype=np.float)
            ksl.append('k3')

    assert len(mbtrs) >= 1


    print('')
    j = 0
    jmax = N_clu*len(ksl)
    strTime = time()

    for i in range(N_clu):
        tmp_mbtr = mbtr.create(trajs[i],n_jobs=8)
        print(tmp_mbtr)
        print(np.shape(tmp_mbtr))

        raise NotImplementedError
        for k in ksl:
            mbtrs[k][i] = tmp_mbtr[k]

            j+=1
            sys.stdout.write('\r({2:.1%}) {0}/{1}'.format(
                    j,jmax,j/jmax))
    endTime = time()
    print('')
    print('Elapsed time for Dscribe MBTR-calculations:')
    print('  Total:   {0:f}'.format(endTime-strTime))
    print('  Total/N: {0:f}'.format((endTime-strTime)/N_clu))
    return mbtrs



def mbtrDataExtraction(k=2, N=-1, flatten=True, acc_N=100, full=False):
    """
    Reads DFT cluster data and converts it to MBTR data as input and
    potential energy as output
    """
    inds = ':'
    if N!=-1:
        inds += '{0:d}'.format(N)
        
    trajPth = Path('Datasets/Au38_QT_all.traj')
    trajs = read(trajPth.as_posix(),index=inds)

    atmsym = set(map(lambda x: atomic_numbers[x.symbol], trajs[0]))
    atomic_symbols = list(sorted(list(atmsym)))
    N_as = len(atomic_symbols)
    if full:
        ks = [1,2,3]
    else:
        ks = k

    # setup dscribe MBTR:
    dmbtr = setup_Dscribe_MBTR(atomic_symbols, ks, acc_N, flatten=flatten)

    X = calculate_Dscribe_mbtrs(trajs, dmbtr, N_as, acc_N, ks,
                                    flatten=flatten)

    Y = list(map(lambda x: x.get_potential_energy(), trajs))
    Y = np.array(Y).reshape(-1,1)

    return (X,Y)


def dsC7O2H10nsd(k=2, N=-1, flatten=True, acc_N=100):
    """
    """
    fldr = Path('Data/dsgdb9nsd/dsC7O2H10nsd')
    xyzs = getFilePaths(fldr, '.xyz')
    xyzs = list(sorted(xyzs))
    #N = len(xyzs)
    traj = []
    for xyz in xyzs:
        traj.append(read(xyz,format='xyz'))
    print(traj)
    


def QM9(k=2, N=-1, flatten=True, acc_N=100):
    """
    https://docs.dgl.ai/generated/dgl.data.QM9Dataset.html#dgl.data.QM9Dataset
    
    {1, 6, 7, 8, 9}
    """
    from dgl.data import QM9Dataset
    
    data = QM9Dataset(label_keys=['U'])#'mu','alpha','homo','lumo'])
    N = len(data)
    
    atmsym = set([1, 6, 7, 8, 9])
    #atmsym = set(map(lambda x: atomic_numbers[x.symbol], ))
    atomic_symbols = list(sorted(list(atmsym)))
    N_as = len(atomic_symbols)
    ks = 2
    #acc_N = 100

    #flatten = True
    # setup dscribe MBTR:
    dmbtr = setup_Dscribe_MBTR(atomic_symbols, ks, acc_N, flatten=flatten)
    
    trajs = []
    Y = np.zeroes(N)
    i = 0
    #Zs = set()
    for g, label in data:
        U = label
        Y[i] = U
        i+=1
        R = g.ndata['R']
        Z = g.ndata['Z']
        syms = list(map(lambda x: chemical_symbols[x], Z))
        #a = Atoms(symbols=syms,positions=R)
        trajs.append(Atoms(symbols=syms,positions=R))
        
    X = calculate_Dscribe_mbtrs(trajs, dmbtr, N_as, acc_N, ks,
                                    flatten=flatten)    
    
    Y = np.array(Y).reshape(-1,1)     
    return (X,Y)


def QM9Alu(letters, k=2, N=-1, flatten=True, acc_N=100, full=False):
    """
    jp6b10908_si_001 supplementary datasets A, B, C, D, and E
    """
    qm9fldr = Path('Datasets/jp6b10908_si_001/supplemental/qm_reference_datafiles')
    
    syms = ['Al']
    datsetFldrs = ['Dataset{0:s}'.format(x) for x in letters]
    #fps = getFilePaths(datsetFldrs, '.xyz')
    xyzs = []
    for dsf in datsetFldrs:
        fps = getFilePaths(Path(qm9fldr,dsf), '.xyz')
        for fp in fps:
            xyzs.extend(read(fp,index=':'))
            
    #N = len(xyzs)  
    if full:
        ks = [1,2,3]
    else:
        ks = k
    dmbtr = setup_Dscribe_MBTR(syms, ks, acc_N, flatten=flatten)
    
    N_as = len(syms)
    X = calculate_Dscribe_mbtrs(xyzs, dmbtr, N_as, acc_N, ks,
                                    flatten=flatten)

    Y = list(map(lambda x: x.get_potential_energy(), xyzs))
    Y = np.array(Y).reshape(-1,1)    
    return (X,Y)


def _readForces(fp: Path) -> np.ndarray:
    """
    """
    lines = []
    with open(fp.as_posix(),'r') as f:
        lines = f.readlines()
        
    lines = lines[slice(2,None)]
    lines = map(lambda x: list(filter(lambda y: len(y)>0,x.split(' ')))[slice(5,8)], lines)
    lines = map(lambda x: list(map(lambda y: float(y), x)), lines)
    lines = map(lambda x: np.linalg.norm(x), lines)
    
    lines = list(lines)
    return np.array(lines)


def SOAP_Alu(letters):
    """
    """
    qm9fldr = Path('Datasets/jp6b10908_si_001/supplemental/qm_reference_datafiles')
    
    syms = ['Al']
    datsetFldrs = ['Dataset{0:s}'.format(x) for x in letters]
    xyzs = []
    Y = []
    for dsf in datsetFldrs:
        fps = getFilePaths(Path(qm9fldr,dsf), '.xyz')
        for fp in fps:
            xyzs.extend(read(fp,index=':',format='extxyz'))
            Y.extend(_readForces(fp))
    
    soap = setup_Dscribe_SOAP(syms, 5, n_max=7, l_max=4, sigma=0.25)
    
    res = soap.create(xyzs,n_jobs=8,verbose=True)
    
    X = []
    for r in res:
        X.extend(r)
    
    return X, Y


def sGDMLconversion(R,z,k=2, N=-1, flatten=True, acc_N=100, full=False):
    """
    Convert sGDML dataset atom coordinates and atom numbers into 
    MBTR descriptors
    """
    syms = list(map(lambda x: chemical_symbols[x], np.sort(np.unique(z))))
    symbols = list(map(lambda x: chemical_symbols[x], z))
    
    
    if full:
        ks = [1,2,3]
    else:
        ks = k
    dmbtr = setup_Dscribe_MBTR(syms, ks, acc_N, flatten=flatten)
    
    atoms = []
    for pos in R:
        atoms.append(Atoms(symbols=symbols,positions=pos))
    
    N_as = len(syms)
    X = calculate_Dscribe_mbtrs(atoms, dmbtr, N_as, acc_N, ks,
                                    flatten=flatten)
    return X


if __name__ == "__main__":
    """
    (X1,Y1) = mbtrDataExtraction(N=1,flatten=True,acc_N=3)
    (X2,Y2) = mbtrDataExtraction(N=1,flatten=False,acc_N=3)
    print('X1:',np.shape(X1))
    print('Y1:',np.shape(Y1))
    print('X2:',np.shape(X2))
    print('Y2:',np.shape(Y2))
    """
    #dsC7O2H10nsd(k=2, N=-1, flatten=True, acc_N=100)
    """
    import matplotlib.pyplot as plt
    
    plt.figure()
    (X,Y) = QM9Alu(['A'],full=False)
    print(X.shape)
    print(Y.shape)
    for i in range(len(X)):
        plt.plot(range(100),X[i])
    
    plt.figure()
    (X,Y) = QM9Alu(['A'],full=True)
    print(X.shape)
    print(Y.shape)
    for i in range(len(X)):
        plt.plot(range(300),X[i])
    """  
        
    #X, Y = SOAP_Alu(['A'])
    
    print('\nEnd of Line')
